""" django_listing root __init__.py file

   isort:skip_file
"""

__version__ = "0.10.28"
__author__ = "Eric Lapouyade"
__copyright__ = "Copyright 2018, The Django listing Project"
__credits__ = ["Eric Lapouyade"]
__license__ = "Dual licensing"
__maintainer__ = "Eric Lapouyade"
__status__ = "Beta"

import re

EXPORT_FORMATS = ["CSV", "DBF", "HTML", "JSON", "ODS", "TSV", "XLS", "XLSX", "YAML"]
EXPORT_FORMATS_KEEP_ORIGINAL_TYPE = ["XLSX", "JSON", "XLS", "DBF"]
EXPORT_FORMATS_USE_COL_NAME = ["JSON"]
EXPORT_EXCEL_SANITIZE_RE = re.compile(r"[\000-\010]|[\013-\014]|[\016-\037]")
FILTER_QUERYSTRING_PREFIX = "f_"

from .listing import *
from .columns import *
from .exceptions import *
from .views import *
from .html_attributes import *
from .context import *
from .paginators import *
from .toolbar import *
from .filters import *
from .aggregations import *
from .attached_form import *
from .actions_buttons_column import *
from .charts import *



#
# Created : 2018-02-06
#
# @author: Eric Lapouyade
#

from django.core.paginator import Paginator as DjangoPaginator, Page as DjangoPage
from django.utils.translation import pgettext_lazy, gettext

from .context import RenderContext
from .theme_config import ThemeAttribute, ThemeTemplate

__all__ = [
    "NoIconButtonPaginator",
    "NoIconButtonPaginatorMixin",
    "NoTextButtonPaginator",
    "NoTextButtonPaginatorMixin",
    "Paginator",
]

PAGINATOR_PARAMS_KEYS = {
    "fast_page_next_tpl",
    "fast_page_prev_tpl",
    "fast_page_step",
    "goto_page_tpl",
    "has_editable_page_info",
    "has_first_last",
    "has_goto_page",
    "has_page_info",
    "has_prev_next",
    "has_row_info",
    "hide_disabled_buttons",
    "hide_single_page",
    "in_footer",
    "page_info_tpl",
    "page_row_tpl",
    "page_scale_ellipsis",
    "page_scale_size",
    "parts_order",
    "template_name",
    "theme_button_a_class",
    "theme_button_li_class",
    "theme_button_text_class",
    "theme_fast_next_icon",
    "theme_fast_page_has_icon",
    "theme_fast_page_has_text",
    "theme_fast_prev_icon",
    "theme_first_icon",
    "theme_first_last_has_icon",
    "theme_first_last_has_text",
    "theme_last_icon",
    "theme_next_icon",
    "theme_prev_icon",
    "theme_prev_next_has_icon",
    "theme_prev_next_has_text",
}


class Paginator(DjangoPaginator):
    template_name = ThemeTemplate("paginator.html")
    has_page_info = True
    has_editable_page_info = False
    page_info_tpl = pgettext_lazy("paginator", "Page {page_number} of {nb_pages}")
    has_row_info = False
    row_info_tpl = pgettext_lazy("paginator", "{row_first}-{row_last} of {nb_rows}")
    has_prev_next = True
    prev_text = pgettext_lazy("paginator", "Previous")
    next_text = pgettext_lazy("paginator", "Next")
    has_first_last = False
    first_text = pgettext_lazy("paginator", "First")
    last_text = pgettext_lazy("paginator", "Last")
    fast_page_step = 0
    fast_page_prev_tpl = "-{step}"
    fast_page_next_tpl = "+{step}"
    page_scale_size = 0
    page_scale_ellipsis = 0
    hide_disabled_buttons = False
    hide_single_page = False
    parts_order = (
        "first,fastprev,prev,pageinfo,rowinfo,scale,next,fastnext,last,gotopage"
    )
    has_goto_page = False
    goto_page_tpl = pgettext_lazy("paginator", "Go to page {goto_form}")
    in_footer = False

    theme_first_last_has_icon = ThemeAttribute("paginator_theme_first_last_has_icon")
    theme_first_last_has_text = ThemeAttribute("paginator_theme_first_last_has_text")
    theme_first_icon = ThemeAttribute("paginator_theme_first_icon")
    theme_last_icon = ThemeAttribute("paginator_theme_last_icon")
    theme_fast_page_has_icon = ThemeAttribute("paginator_theme_fast_page_has_icon")
    theme_fast_page_has_text = ThemeAttribute("paginator_theme_fast_page_has_text")
    theme_fast_prev_icon = ThemeAttribute("paginator_theme_fast_prev_icon")
    theme_fast_next_icon = ThemeAttribute("paginator_theme_fast_next_icon")
    theme_prev_next_has_icon = ThemeAttribute("paginator_theme_prev_next_has_icon")
    theme_prev_next_has_text = ThemeAttribute("paginator_theme_prev_next_has_text")
    theme_prev_icon = ThemeAttribute("paginator_theme_prev_icon")
    theme_next_icon = ThemeAttribute("paginator_theme_next_icon")
    theme_button_a_class = ThemeAttribute("paginator_theme_button_a_class")
    theme_button_li_class = ThemeAttribute("paginator_theme_button_li_class")
    theme_button_text_class = ThemeAttribute("paginator_theme_button_text_class")

    def __init__(
        self, listing, object_list, per_page, orphans=0, allow_empty_first_page=True
    ):
        super().__init__(object_list, per_page, orphans, allow_empty_first_page)
        self.listing = listing
        for k in PAGINATOR_PARAMS_KEYS:
            listing_key = "paginator_" + k
            if hasattr(listing, listing_key):
                setattr(self, k, getattr(listing, listing_key))
        if isinstance(self.parts_order, str):
            self.parts_order = self.parts_order.replace(" ", "")
            self.parts_order = list(
                map(lambda s: s.split(","), self.parts_order.split(";"))
            )

    def validate_number(self, number):
        if number == "last":
            number = self.num_pages
        return super().validate_number(number)

    def page(self, number):
        """Return a Page object for the given 1-based page number."""
        number = self.validate_number(number)
        bottom = (number - 1) * self.per_page
        top = bottom + self.per_page
        if bottom > self.count - self.per_page:
            bottom = max(0, self.count - self.per_page)
            top = self.count
        if bottom > self.listing.offset_max:
            self.listing.empty_table_msg = (
                gettext(
                    "The listing can only display the first %s rows. "
                    "Please use filtering to narrow your listing."
                )
                % self.listing.offset_max
            )
            return Page([], number, self, bottom, top)
        page = Page(self.object_list[bottom:top], number, self, bottom, top)
        return page

    def get_context(self):
        get_url = self.listing.get_url
        page = self.listing.current_page
        fast_page_prev = max(1, page.number - self.fast_page_step)
        fast_page_next = min(self.num_pages, page.number + self.fast_page_step)
        beginning_ellipsis_pages = []
        beginning_ellipsis_display = True
        ending_ellipsis_pages = []
        ending_ellipsis_display = True
        if self.page_scale_size:
            scale_range_min = max(1, page.number - int(self.page_scale_size / 2))
            scale_range_max = min(
                scale_range_min + self.page_scale_size - 1, self.num_pages
            )

            if self.page_scale_ellipsis:
                beginning_ellipsis_pages = [
                    (p, get_url(page=p), p == page.number)
                    for p in range(
                        1, min(self.page_scale_ellipsis + 1, scale_range_min)
                    )
                ]
                if scale_range_min < self.page_scale_ellipsis + 2:
                    beginning_ellipsis_display = False
                if scale_range_max == self.num_pages:
                    scale_range_min = max(1, self.num_pages - self.page_scale_size)
                else:
                    ending_ellipsis_range_min = max(
                        scale_range_max + 1,
                        self.num_pages - self.page_scale_ellipsis + 1,
                    )
                    if ending_ellipsis_range_min == scale_range_max + 1:
                        beginning_ellipsis_display = False
                    ending_ellipsis_range_max = self.num_pages
                    ending_ellipsis_pages = [
                        (p, get_url(page=p), p == page.number)
                        for p in range(
                            ending_ellipsis_range_min, ending_ellipsis_range_max + 1
                        )
                    ]

            scale_pages = [
                (p, get_url(page=p), p == page.number)
                for p in range(scale_range_min, scale_range_max + 1)
            ]
        else:
            scale_pages = []

        if self.has_editable_page_info or self.has_goto_page:
            goto_form = '<form class="goto-page">'
            goto_form += self.listing.get_hiddens_html(without="page")
            goto_form += (
                '<input type="text" '
                'value="{val}" '
                'name="page{suffix}">'.format(
                    val=page.number, suffix=self.listing.suffix
                )
            )
            goto_form += "</form>"
        else:
            goto_form = ""

        if self.has_editable_page_info:
            page_number = goto_form
        else:
            page_number = page.number

        nb_page_rows = page.end_index() - page.start_index() + 1

        return RenderContext(
            first_page_url=get_url(page=1),
            last_page_url=get_url(page=self.num_pages),
            prev_page_url=get_url(
                page=page.previous_page_number() if page.number > 1 else 1
            ),
            next_page_url=get_url(
                page=page.next_page_number()
                if page.number < self.num_pages
                else self.num_pages
            ),
            fast_page_prev_url=get_url(page=fast_page_prev),
            fast_page_next_url=get_url(page=fast_page_next),
            fast_page_prev_text=self.fast_page_prev_tpl.format(
                step=self.fast_page_step
            ),
            fast_page_next_text=self.fast_page_next_tpl.format(
                step=self.fast_page_step
            ),
            scale_pages=scale_pages,
            beginning_ellipsis_pages=beginning_ellipsis_pages,
            beginning_ellipsis_display=beginning_ellipsis_display,
            ending_ellipsis_pages=ending_ellipsis_pages,
            ending_ellipsis_display=ending_ellipsis_display,
            nb_page_rows=nb_page_rows,
            page_info=self.page_info_tpl.format(
                page_number=page_number, nb_pages=self.num_pages
            ),
            row_info=self.row_info_tpl.format(
                row_first=page.start_index(),
                row_last=page.end_index(),
                nb_rows=self.count,
                nb_page_rows=nb_page_rows,
            ),
            goto_page=self.goto_page_tpl.format(goto_form=goto_form),
            show_paginator_single_page=(
                self.num_pages > 1 or not self.hide_single_page
            ),
            current_page=page,
            paginator=self,
        )


class Page(DjangoPage):
    def __init__(self, object_list, number, paginator, bottom=None, top=None):
        super().__init__(object_list, number, paginator)
        self._bottom = bottom
        self._top = top

    def start_index(self):
        """
        Return the 1-based index of the first object on this page,
        relative to total objects in the paginator.
        """
        # Special case, return zero if no items.
        if self.paginator.count == 0:
            return 0
        # Special case for django-listing because last page need special calculation
        if self._bottom is not None:
            return self._bottom + 1
        return (self.paginator.per_page * (self.number - 1)) + 1

    def end_index(self):
        """
        Return the 1-based index of the last object on this page,
        relative to total objects found (hits).
        """
        # Special case for django-listing because last page need special calculation
        if self._top is not None:
            return self._top
        # Special case for the last page because there can be orphans.
        if self.number == self.paginator.num_pages:
            return self.paginator.count
        return self.number * self.paginator.per_page


class NoTextButtonPaginatorMixin:
    theme_first_last_has_text = False
    theme_fast_page_has_text = False
    theme_prev_next_has_text = False


class NoTextButtonPaginator(NoTextButtonPaginatorMixin, Paginator):
    pass


class NoIconButtonPaginatorMixin:
    theme_first_last_has_icon = False
    theme_fast_page_has_icon = False
    theme_prev_next_has_icon = False


class NoIconButtonPaginator(NoIconButtonPaginatorMixin, Paginator):
    pass



#
# Created : 2018-02-10
#
# @author: Eric Lapouyade
#

from django.utils.safestring import mark_safe

__all__ = ["HTMLAttributes", "Tag", "Html"]


class HTMLAttributes(dict):
    def __init__(self, *args, **kwargs):
        # remove starting underscore to be able to use '_class'
        # instead of 'class' in kwargs (reserved keyword)
        if "_class" in kwargs:
            kwargs["class"] = kwargs.pop("_class")
        super().__init__(*args, **kwargs)

    def add(self, attr, value):
        if value is not None:
            if isinstance(value, int):
                value = str(value)
            if attr not in self:
                s = set()
            else:
                if attr == "style":
                    s = set(self[attr].split(";"))
                else:
                    s = set(self[attr].split())
            if isinstance(value, set):
                s.update(value)
            else:
                s.add(value)
            if attr == "style":
                # use str to consume lazy strings
                self[attr] = (";".join(map(str, s))).strip()
            else:
                # use str to consume lazy strings
                self[attr] = (" ".join(map(str, s))).strip()

    def remove(self, attr, value):
        if value is not None:
            if attr not in self:
                s = set()
            else:
                if attr == "style":
                    s = set(self[attr].split(";"))
                else:
                    s = set(self[attr].split())
            if isinstance(value, set):
                s -= value
            else:
                s -= set([value])  # do not use remove() to avoid KeyError
            if attr == "style":
                self[attr] = (";".join(s)).strip()
            else:
                self[attr] = (" ".join(s)).strip()

    def set(self, attr, value=None):
        self[attr] = value

    def __setattr__(self, attr, value):
        self[attr] = value

    def update(self, *args, **kwargs):
        # remove starting underscore to be able to use '_class'
        # instead of 'class' in kwargs (reserved keyword)
        if "_class" in kwargs:
            kwargs["class"] = kwargs.pop("_class")
        super().update(*args, **kwargs)

    def copy(self):
        return HTMLAttributes(self)

    def __str__(self):
        out = " ".join(
            [(k if v is None else '{}="{}"'.format(k, v)) for k, v in self.items()]
        )
        if out:
            out = " " + out
        return mark_safe(out)


class Tag:
    def __init__(self, tag, text, **kwargs):
        """If named argument begins by a "_" replace them by a "-"
        this is useful because, "-" cannot be used in a name in python.
        This is also useful to use class which is reserved : use "_class" instead.
        """
        attributes = {}
        for k, v in kwargs.items():
            if k.startswith("_"):
                k = k.replace("_", "-")[1:]
            attributes[k] = v
        self.tag = tag
        self.text = text
        self.attributes = attributes

    def __str__(self):
        attributes_html = " ".join(f'{k}="{v}"' for k, v in self.attributes.items())
        return mark_safe(f"<{self.tag} {attributes_html}>{self.text}</{self.tag}>")


class Html:
    def __init__(self, html):
        self.html = html

    def __str__(self):
        return mark_safe(self.html)



from django_listing.theme_config import ThemeTemplate
from decimal import Decimal
from django.conf import settings
from .utils import to_js_timestamp

__all__ = [
    "BaseChartMixin",
    "PieChartMixin",
    "BarChartMixin",
    "TimestampedBarChartMixin",
    "TimestampedLineChartMixin",
]


class BaseChartMixin:
    listing_template_name = ThemeTemplate("chart.html")
    chart_height = 600
    chart_width = 800

    def get_chart_records(self):
        # If you derive this method, you need to cache results as this method
        # will be called many times.
        # Note : self.records.current_page() already caches records
        return self.records.current_page()

    def get_chart_rec_color(self, rec):
        idx = rec.get_index()
        dlst_settings = settings.django_listing_settings
        gradient = dlst_settings.CHARTS_DEFAULT_GRADIENT
        gradient_default = dlst_settings.CHARTS_DEFAULT_GRADIENT_OVERFLOW
        color = gradient[idx] if idx < len(gradient) else gradient_default
        return color

    def get_chart_rec_values(self, rec):
        val = rec.get(self.chart_values_rec_key)
        if isinstance(val, Decimal):
            val = float(val)
        return val

    def get_chart_rec_label(self, rec):
        if self.chart_labels_rec_key is None:
            return None
        return rec.get(self.chart_labels_rec_key, "-")

    def get_chart_height(self):
        return self.chart_height

    def get_chart_width(self):
        return self.chart_width

    def get_apex_options(self):
        raise NotImplemented

    def get_chart_json_id(self):
        return f"{self.css_id}-chart-json-id"

    def get_chart_div_id(self):
        return f"{self.css_id}-chart-id"

    def chart_data(self):
        chart_div_id = self.get_chart_div_id()
        return dict(
            chart_div_id=self.get_chart_div_id(),
            json_id=self.get_chart_json_id(),
            apexcharts_options=self.get_apex_options(
                chart_div_id, *self.get_chart_args()
            ),
        )

    def get_chart_args(self):
        colors = []
        values = []
        labels = []
        for rec in self.get_chart_records():
            colors.append(self.get_chart_rec_color(rec))
            values.append(self.get_chart_rec_values(rec))
            labels.append(self.get_chart_rec_label(rec))
        width = self.get_chart_width()
        height = self.get_chart_height()
        return values, labels, colors, width, height


class PieChartMixin(BaseChartMixin):
    listing_template_name = ThemeTemplate("chart_pie.html")

    def get_apex_options(self, chart_div_id, values, labels, colors, width, height):
        options = {
            "series": values,
            "chart": {
                "id": chart_div_id,
                "width": width,
                "height": height,
                "type": "pie",
            },
            "colors": colors,
            "labels": labels,
            "dataLabels": {"style": {"fontSize": "16px"}},
        }
        return options


class BarChartMixin(BaseChartMixin):
    listing_template_name = ThemeTemplate("chart_bar.html")

    def get_apex_options(self, chart_div_id, values, labels, colors, width, height):
        options = {
            "series": [{"data": values, "name": ""}],
            "chart": {
                "id": chart_div_id,
                "width": width,
                "height": height,
                "type": "bar",
            },
            "colors": colors,
            "plotOptions": {
                "bar": {
                    "columnWidth": "45%",
                    "distributed": True,
                    "dataLabels": {
                        "position": "top",
                    },
                }
            },
            "dataLabels": {
                "enabled": True,
                "offsetY": -20,
                "style": {"fontSize": "12px", "colors": ["#222222"]},
            },
            "legend": {"show": False},
            "xaxis": {
                "categories": labels,
                "labels": {"style": {"colors": colors, "fontSize": "12px"}},
            },
        }
        return options


class TimestampedChartMixin(BaseChartMixin):
    chart_labels_rec_key = None

    def get_chart_rec_color(self, rec):
        return settings.django_listing_settings.CHARTS_DEFAULT_TREND_BAR_COLOR

    def get_chart_rec_values(self, rec):
        val = rec.get(self.chart_values_rec_key)
        timestamp = to_js_timestamp(rec.get(self.chart_timestamps_rec_key))
        if isinstance(val, Decimal):
            val = float(val)
        return [timestamp, val]


class TimestampedBarChartMixin(TimestampedChartMixin):
    listing_template_name = ThemeTemplate("chart_trend_bar.html")

    def get_apex_options(self, chart_div_id, values, labels, colors, width, height):
        options = {
            "colors": colors,
            "series": [
                {
                    "name": "Active values (Sum)",
                    "data": values,
                }
            ],
            "chart": {
                "id": chart_div_id,
                "type": "bar",
                "width": width,
                "height": height,
                "zoom": {
                    "autoScaleYaxis": True,
                },
            },
            "plotOptions": {
                "bar": {
                    "columnWidth": "85%",
                }
            },
            "dataLabels": {
                "enabled": False,
            },
            "markers": {
                "size": 0,
                "style": "hollow",
            },
            "xaxis": {
                "type": "datetime",
                "tickAmount": 6,
            },
            "tooltip": {
                "x": {
                    "format": "dd MMM yyyy",
                }
            },
        }
        return options


class TimestampedLineChartMixin(BaseChartMixin):
    listing_template_name = ThemeTemplate("chart_trend_bar.html")
    chart_labels_rec_key = None
    per_page = 1000

    def get_chart_rec_color(self, rec):
        return settings.django_listing_settings.CHARTS_DEFAULT_TREND_BAR_COLOR

    def get_chart_rec_values(self, rec):
        val = rec.get(self.chart_values_rec_key)
        timestamp = rec.get(self.chart_timestamps_rec_key)
        if isinstance(val, Decimal):
            val = float(val)
        return [timestamp, val]

    def get_apex_options(self, chart_div_id, values, labels, colors, width, height):
        options = {
            "colors": colors,
            "series": [
                {
                    "name": "Active values (Sum)",
                    "data": values,
                }
            ],
            "chart": {
                "id": "chart",
                "type": "line",
                "width": width,
                "height": height,
                "zoom": {
                    "autoScaleYaxis": True,
                },
            },
            "dataLabels": {
                "enabled": False,
            },
            "markers": {
                "size": 0,
                "style": "hollow",
            },
            "xaxis": {
                "type": "datetime",
                "tickAmount": 6,
            },
            "tooltip": {
                "x": {
                    "format": "dd MMM yyyy",
                }
            },
        }
        return options



#
# Created : 2018-02-09
#
# @author: Eric Lapouyade
#
import traceback
from urllib.parse import parse_qs

from django import forms
from django.contrib import messages
from django.core.exceptions import PermissionDenied

from django.db import models
from django.db.models import QuerySet
from django.forms.models import construct_instance
from django.http import (
    HttpResponse,
    HttpResponseRedirect,
    HttpResponseServerError,
    QueryDict,
    JsonResponse,
)
from django.template import RequestContext, loader
from django.utils.module_loading import import_string
from django.utils.safestring import mark_safe
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _
from django.views.generic import TemplateView

from .exceptions import *
from .listing import Listing, logger
from .attached_form import AttachedForm

__all__ = [
    "INSTANCE_METHOD_PREFIX",
    "LISTING_REDIRECT_NONE",
    "LISTING_REDIRECT_NO_EDIT",
    "LISTING_REDIRECT_SAME_PAGE",
    "ListingView",
    "ListingViewMixin",
]

from .utils import is_ajax

INSTANCE_METHOD_PREFIX = "get_listing_instance_"
LISTING_REDIRECT_NONE = None
LISTING_REDIRECT_SAME_PAGE = 1
LISTING_REDIRECT_NO_EDIT = 2


class ListingViewMixin:
    listing_class = None
    listing_data = None
    upload_field = "image"
    context_classes = ()
    listing_context_name = "listing"
    listing_instance = None
    listing = None
    insert_success_redirect_url = LISTING_REDIRECT_NONE
    insert_success_msg = _("<b>{object}</b> has been successfully added.")
    insert_success_msg_no_save = _(
        "The form is valid but nothing has been added to database "
        "as <tt>save_to_database=False</tt>."
    )
    update_success_redirect_url = LISTING_REDIRECT_NONE
    update_success_msg = _(
        "<b>{nb_updates} {model_verbose}</b> " "has been successfully updated."
    )
    update_success_msg_no_save = _(
        "The form is valid but nothing has been updated to database "
        "as <tt>save_to_database=False</tt>."
    )
    is_ajax = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._listing_instances = {}
        self._formset_errors = {}
        self.ajax_request_context = None

    def post(self, request, *args, **kwargs):
        # make POST data mutable
        request.POST = request.POST.copy()
        if hasattr(self, "get_object"):  # for DetailView that works only on GET
            self.object = self.get_object()
        try:
            self.is_ajax = is_ajax(request)
            if self.is_ajax:
                try:
                    if "serialized_data" in request.POST:
                        serialized_data = request.POST.pop("serialized_data")
                        if isinstance(serialized_data, list):
                            serialized_data = serialized_data[0]
                        data = parse_qs(serialized_data)
                        for k, v in data.items():
                            if k != "csrfmiddlewaretoken":
                                if len(v) == 1:
                                    request.POST[k] = v[0]
                                else:
                                    request.POST[k] = v
                        # Manage checkbox cases on mass-op :
                        # When checkbox is off : JS serialize do not transmit checkbox
                        # input name, so create it...
                        for k in data.keys():
                            if k.endswith("_mass_op"):
                                input_name = k[: -len("_mass_op")]
                                if input_name not in request.POST:
                                    request.POST[input_name] = ""

                    return self.manage_listing_ajax_request(request, *args, **kwargs)
                except KeyboardInterrupt:
                    raise
                except Exception as e:
                    stack = traceback.format_exc()
                    logger.error(stack)
                    return HttpResponseServerError(e)
            response = self.manage_listing_post(request, *args, **kwargs)
            if response:
                return response
            return self.get(request, *args, **kwargs)
        except ListingException as e:
            return HttpResponseServerError(str(e))

    def get(self, request, *args, **kwargs):
        response = super().get(request, *args, **kwargs)
        # need to force rendering here to know whether a listing created
        # in a template has requested a data export
        response.render()
        if hasattr(request, "export_data"):
            data = request.export_data
            filename = getattr(request, "export_filename", "listing")
            response = HttpResponse(data)
            response["Content-Disposition"] = 'attachment; filename="{}"'.format(
                filename
            )
            response.set_cookie("file_generation", "done")
            response.set_cookie("last_file_generation_filename", filename)
            return response
        return response

    def json_response(self, data, **kwargs):
        if hasattr(self, "listing_patch_json_response_data"):
            self.listing_patch_json_response_data(data)
        response = JsonResponse(data, **kwargs)
        response["Cache-Control"] = "no-cache"
        return response

    def get_listing_from_post(self, request, refresh=False):
        listing_id = request.POST.get("listing_id", "")[:-3]  # remove '-id' suffix
        listing_suffix = request.POST.get("listing_suffix", "")
        listing = self.get_listing_id(listing_id, refresh)
        if not listing:
            raise ListingException(
                "{}-id is a bad listing ID : django-listing library misconf"
                "iguration. Contact the webmaster !".format(listing_id)
            )
        Listing.set_suffix(request, listing, listing_suffix)
        listing.action = request.POST.get("force_action") or request.POST.get("action")
        listing.action_col = request.POST.get("action_col")
        listing.request = request
        return listing

    def manage_listing_ajax_request(self, request, *args, **kwargs):
        listing = self.get_listing_from_post(request)
        self.ajax_request_context = RequestContext(request)
        self.ajax_request_context.update(self.get_context_data(**kwargs))
        self.listing = listing
        response = None
        if listing:
            listing.set_view(self)
            listing_part = request.POST.get("listing_part", "all")
            listing.ajax_request = True
            listing.ajax_part = listing_part
            if isinstance(listing.action, str) and listing.action:
                listing.render_init_context(self.ajax_request_context)
                method = getattr(self, "manage_listing_%s" % listing.action, None)
                if callable(method):
                    response = method(listing, *args, **kwargs)
        if response:
            return response
        if listing.have_to_refresh():
            listing = self.get_listing_from_post(request, refresh=True)
        response_data = {}
        listing.render_init(self.ajax_request_context)
        if filters := getattr(listing, "filters", None):
            cleaned_data = filters.get_cleaned_data()
            if cleaned_data is None:
                listing.filters.form().initial.update(redraw_filters_form=True)
            # If filters form is invalid or was the previous time : redraw it
            if (
                cleaned_data is None
                or "redraw_filters_form" in listing.request_data
                or listing.request_data.get("filters_action") == "reset"
            ):
                filters_form_html = filters.render_form(self.ajax_request_context)
                response_data["filters_form"] = filters_form_html
            if cleaned_data is None:  # filter form is invalid
                # return only filters form
                return JsonResponse(response_data)
        response_data["listing"] = listing.render(self.ajax_request_context)
        return JsonResponse(response_data)

    def get_listing_class(self):
        return self.listing_class

    def get_listing_data(self):
        return self.listing_data

    def get_listing_context_name(self):
        return self.listing_context_name

    def get_listing_params(self):
        keys = list(Listing.get_params_keys()) + [
            k for k in self.__class__.__dict__ if "__" in k and not k.startswith("__")
        ]  # add listing columns custimization keys (colname__attribute)
        return {k: getattr(self, k) for k in keys if hasattr(self, k)}

    def get_listing_instance(self):
        data = self.get_listing_data()
        if isinstance(data, (QuerySet, list)) or data:
            listing_class = self.get_listing_class() or Listing
            return listing_class(data, **self.get_listing_params())

    # this set of methods manage a special storage because when using ajax
    # listing name are taken from div id which has been normalized
    # by replacing underscores by dashes
    def get_from_listing_instances(self, name):
        v = self._listing_instances.get(name.replace("_", "-"))
        if v:
            return v[1]
        return None

    def in_listing_instances(self, name):
        return name.replace("_", "-") in self._listing_instances

    def set_to_listing_instances(self, name, instance):
        self._listing_instances[name.replace("_", "-")] = (name, instance)

    def yield_listing_instances(self):
        for v in self._listing_instances.values():
            yield v[1]

    def listing_instances_context(self):
        return dict(self._listing_instances.values())

    def get_default_listing_instance(self):
        context_name = self.get_listing_context_name()
        instance = self.get_from_listing_instances(context_name)
        if not instance:
            instance = self.get_listing_instance()
            if instance:
                # instance.id = (
                #     context_name + "-id"
                # )  # Ensure id is set according to method name + '-id'
                self.set_to_listing_instances(context_name, instance)
        return instance

    def get_listing_update_success_redirect_url(self, listing=None):
        url = self.update_success_redirect_url
        if listing:
            if url == LISTING_REDIRECT_SAME_PAGE:
                url = listing.get_url()
            elif url == LISTING_REDIRECT_NO_EDIT:
                url = listing.get_url(without="editing,editing_columns,editing_row_pk")
        # if url = None will do a LISTING_REDIRECT_NONE : the post() will call get()
        # at the end on the same request.
        return url

    def get_listing_insert_success_redirect_url(self, listing=None):
        url = self.insert_success_redirect_url
        if listing:
            if url == LISTING_REDIRECT_SAME_PAGE:
                url = listing.get_url()
            elif url == LISTING_REDIRECT_NONE:
                # post() will call get() on-the-fly
                # to avoid a new request to the server.
                # That requires to reset insert form by removing Post data
                listing.request.POST = QueryDict()
        return url

    def manage_listing_post(self, request, *args, **kwargs):
        if "action" in request.POST:
            listing = self.get_listing_from_post(request)
            self.listing = listing
            response = None
            if listing:
                listing.set_view(self)
                listing.render_init(RequestContext(request))
                if isinstance(listing.action, str) and listing.action:
                    method = getattr(self, "manage_listing_%s" % listing.action, None)
                    if callable(method):
                        response = method(listing, *args, **kwargs)
                # elif listing.can_edit:
                #     response = self.manage_listing_update(listing, *args, **kwargs)
                # elif listing.can_select:
                #     response = self.manage_listing_select(listing, *args, **kwargs)
            return response

    def manage_listing_upload(self, listing, *args, **kwargs):
        # Work in progress !
        listing.model.objects.create(
            **{self.upload_field: listing.request.FILES["file"]}
        )
        return self.json_response(
            dict(
                message="OK",
            )
        )

    def manage_listing_update(self, listing, *args, **kwargs):
        if listing.editable and listing.editing:
            formset = listing.get_formset()
            if formset.is_valid():
                return self.manage_listing_update_valid(listing, formset)
            else:
                # extract error strings from Django ErrorDict
                listing.row_form_errors = [
                    ", ".join(list(e["__all__"]))
                    for e in formset.errors
                    if "__all__" in e
                ]

    def get_form_instance(self, listing):
        if listing.attached_form:
            if isinstance(listing.attached_form, AttachedForm):
                django_form = listing.attached_form.get_form()
            elif isinstance(listing.attached_form, forms.Form):
                django_form = listing.attached_form
            else:
                if isinstance(listing.attached_form, str):
                    form_class = import_string(listing.attached_form)
                else:
                    form_class = listing.attached_form
                django_form = form_class(listing.request.POST, listing.request.FILES)
        else:
            layout = listing.request.POST.get("attached_form_layout")
            layout_name = listing.request.POST.get("attached_form_layout_name")
            name = listing.request.POST.get("attached_form_name")
            if not layout or not name:
                raise InvalidAttachedForm(
                    "At least a form layout and name are mandatory in POST data "
                    "to build a relevant form instance"
                )
            layout_key = f"layout_{layout_name}" if layout_name else "layout"
            kwargs = {
                "name": name,
                layout_key: layout,
                "layout_name": layout_name,
            }
            attached_form = AttachedForm(**kwargs)
            attached_form = attached_form.bind_to_listing(listing)
            django_form = attached_form.get_form()
        return django_form

    def manage_listing_insert(self, listing, *args, **kwargs):
        form = self.get_form_instance(listing)
        if form.is_valid():
            return self.manage_listing_insert_valid(listing, form)

    def manage_listing_action_button(self, listing, *args, **kwargs):
        col_name = listing.action_col
        if col_name:
            col = listing.columns.get(col_name)
            if col:
                return col.manage_button_action(*args, **kwargs)
        raise ListingException(f'Unknown action column "{col_name}"')

    def manage_listing_attached_form(self, listing, *args, **kwargs):
        """
        To manage a simple action that does not required the form to be instanciated,
        one can define such method:

        in view :
            manage_attached_form_<action>_action(self, listing, *args, **kwargs)
        in listing :
            manage_attached_form_<action>_action(self, *args, **kwargs)

        But for form processing, one can define such method:

        in view :
            manage_attached_form_<action>_process(self, listing, form, instance, *args, **kwargs)
        in listing :
            manage_attached_form_<action>_process(self, form, instance, *args, **kwargs)

        For ajax requests methods must return None or a dict like this :

            {
                "listing": <html code to replace actual listing>,
                "attached_form": <html code to replace actual form>,
            }

            If givent listing or attached_form html code is None, it will be
            automatically calculated.
            If methods returns None instead of a the dict. The dict will be
            automatically calculated.
        """
        # Test permissions
        # Note : action_button is automatically set from POST during listing init
        action = listing.action_button
        if not listing.has_permission_for_action(action):
            raise PermissionDenied(gettext("You do not have the required permission"))

        # Read selected listing rows
        selected_rows = listing.get_selected_rows()

        # if a specific action method is available in view or listing : use it
        action_meth_name = f"manage_attached_form_{action}_action"
        # Search method in view object first
        action_method = getattr(self, action_meth_name, None)
        if action_method:
            return action_method(listing, *args, **kwargs)
        action_method = getattr(listing, action_meth_name, None)
        if action_method:
            return action_method(*args, **kwargs)

        # if a specific get_form() method is available in view or listing : use it
        action_meth_name = f"manage_attached_form_{action}_get_form"
        # Search method in listing object first
        get_form_method = getattr(listing, action_meth_name, None)
        if get_form_method:
            form = get_form_method(*args, **kwargs)
        else:
            get_form_method = getattr(self, action_meth_name, None)
            if get_form_method:
                form = get_form_method(listing, *args, **kwargs)
            else:
                attached_form = listing.attached_form
                form = attached_form.get_form()

        # create instance from database if possible
        instance = None
        object_pk = listing.request.POST.get("object_pk")
        if object_pk:
            instance = listing.model.objects.filter(pk=object_pk).first()
        form.instance = instance
        if form.is_valid():
            # if no instance, build it from form cleaned_data:
            if not instance:
                instance = listing.model()
            form.instance = construct_instance(form, instance)

            process_meth_name = f"manage_attached_form_{action}_process"
            # Search method in listing object first
            process_method = getattr(listing, process_meth_name, None)
            if process_method:
                mixed_response = process_method(form, instance, *args, **kwargs)
            else:
                # then in view if not found in listing
                process_method = getattr(self, process_meth_name, None)
                if not process_method:
                    raise ListingException(
                        f'Do not know how to manage "{action}" action. '
                        f"Please define {action_meth_name}() or {process_meth_name}() "
                        f"in your view or your in your listing."
                    )
                mixed_response = process_method(
                    listing, form, instance, *args, **kwargs
                )
            if (
                isinstance(mixed_response, dict)
                and "attached_form" in mixed_response
                and not mixed_response.get("layout_name")
            ):
                mixed_response["layout_name"] = listing.attached_form.layout_name or ""
            if getattr(form, "errors", False):
                raise ListingException(
                    f"You cannot add errors in form object inside "
                    f"{process_method.__qualname__}().\n"
                    f"This is has to be done inside a xxx_clean() method."
                )
            self.listing.request.POST[instance._meta.pk.attname] = instance.pk
            if listing.processed_flash:
                if len(selected_rows) < 2:
                    if instance.pk:
                        listing.processed_pks = {instance.pk}
                else:
                    listing.processed_pks = set(selected_rows)
            if not self.is_ajax:
                return None
            listing.compute_current_page_records()
            attached_form_html = listing.attached_form.render(self.ajax_request_context)
            if mixed_response is None:
                listing_html = listing.render(self.ajax_request_context)
                mixed_response = {
                    "listing": listing_html,
                    "attached_form": attached_form_html,
                    "object_pk": instance.pk,
                }
            else:
                if "listing" in mixed_response and mixed_response["listing"] is None:
                    listing_html = listing.render(self.ajax_request_context)
                    mixed_response["listing"] = listing_html
                if (
                    "attached_form" in mixed_response
                    and mixed_response["attached_form"] is None
                ):
                    mixed_response["attached_form"] = attached_form_html
            if instance.pk:
                mixed_response["object_pk"] = instance.pk
            return self.json_response(mixed_response)
        else:
            if not self.is_ajax:
                return None
            attached_form_html = listing.attached_form.render(self.ajax_request_context)
            return self.json_response({"attached_form": attached_form_html})

    def manage_attached_form_insert_get_form(self, listing, *args, **kwargs):
        form = listing.attached_form.get_form()
        id_field = form.fields.get("id")
        if id_field:
            id_field.required = False
        return form

    def manage_attached_form_insert_process(
        self, listing, form, instance, *args, **kwargs
    ):
        instance.pk = None  # needed to force insert vs update
        instance.save()
        listing.page = "last"
        listing.sort = "id"

    def manage_attached_form_update_get_form(self, listing, *args, **kwargs):
        # Fields are not required only in mass update (more than one row selected)
        form = listing.attached_form.get_form(
            force_not_required=len(listing.get_selected_rows()) > 1
        )
        # trigger default checks
        form.full_clean()
        # add error if no row selected
        if len(listing.get_selected_rows()) == 0:
            form.add_error(
                None, gettext("Please select at least one item in the listing")
            )
        return form

    def manage_attached_form_update_process(
        self, listing, form, instance, *args, **kwargs
    ):
        selected_rows = listing.get_selected_rows()
        if len(selected_rows) == 1 and instance.pk:
            instance.save()
        else:
            update_fields = {
                k: v
                for k, v in form.cleaned_data.items()
                if self.request.POST.get(f"{k}_mass_op") == "checked" and k != "id"
            }
            listing.model.objects.filter(pk__in=selected_rows).update(**update_fields)

    def manage_attached_form_update_all_get_form(self, listing, *args, **kwargs):
        return listing.attached_form.get_form(force_not_required=True)

    def manage_attached_form_update_all_process(
        self, listing, form, instance, *args, **kwargs
    ):
        update_fields = {
            k: v
            for k, v in form.cleaned_data.items()
            if self.request.POST.get(f"{k}_mass_op") == "checked" and k != "id"
        }
        listing.records.get_filtered_queryset().update(**update_fields)

    def manage_attached_form_duplicate_get_form(self, listing, *args, **kwargs):
        form = listing.attached_form.get_form(force_not_required=True)
        selected_rows = listing.get_selected_rows()
        if len(selected_rows) == 0:
            form.add_error(None, gettext("Please select one item"))
        elif len(selected_rows) > 1:
            form.add_error(
                None, gettext("You can duplicate only one item at the same time")
            )
        return form

    def manage_attached_form_duplicate_process(
        self, listing, form, instance, *args, **kwargs
    ):
        if len(listing.get_selected_rows()) == 1 and instance.pk:
            instance.pk = None
            instance.save()

    def manage_attached_form_delete_get_form(self, listing, *args, **kwargs):
        form = listing.attached_form.get_form(do_not_clean=True)
        if len(listing.get_selected_rows()) == 0:
            form.add_error(None, gettext("Please select at least one item"))
        return form

    def manage_attached_form_delete_process(
        self, listing, form, instance, *args, **kwargs
    ):
        selected_rows = listing.get_selected_rows()
        if len(selected_rows) == 1 and instance.pk:
            instance.delete()
        else:
            listing.model.objects.filter(pk__in=selected_rows).delete()

    def manage_attached_form_delete_all_get_form(self, listing, *args, **kwargs):
        return listing.attached_form.get_form(do_not_clean=True)

    def manage_attached_form_delete_all_process(
        self, listing, form, instance, *args, **kwargs
    ):
        listing.records.get_filtered_queryset().delete()

    def manage_attached_form_clear_action(self, listing, *args, **kwargs):
        attached_form = listing.attached_form
        attached_form.set_layout("")
        form_html = attached_form.render(self.ajax_request_context)
        return self.json_response({"attached_form": form_html})

    def listing_save_rows_to_database(self, listing, formset):
        updated_rows_pk = []
        if issubclass(listing.model, models.Model):
            for row in formset.cleaned_data:
                pk = row.get(listing.primary_key)
                if pk:
                    listing.model.objects.filter(pk=pk).update(**row)
                    updated_rows_pk.append(pk)
        return updated_rows_pk

    def manage_listing_update_valid(self, listing, formset):
        if listing.save_to_database:
            updated_rows_pk = self.listing_save_rows_to_database(listing, formset)
        self.send_listing_update_success_message(listing, updated_rows_pk)
        redirect_to = self.get_listing_update_success_redirect_url(listing)
        if redirect_to:
            return HttpResponseRedirect(redirect_to)

    def listing_insert_into_database(self, listing, form):
        if issubclass(listing.model, models.Model):
            object = listing.model(**form.cleaned_data)
            object.save()
        return object

    def manage_listing_insert_valid(self, listing, form):
        object = None
        if listing.save_to_database:
            object = self.listing_insert_into_database(listing, form)
        self.send_listing_insert_success_message(listing, form, object)
        redirect_to = self.get_listing_insert_success_redirect_url(listing)
        if redirect_to:
            return HttpResponseRedirect(redirect_to)

    def send_listing_insert_success_message(self, listing, form, object):
        if listing.save_to_database:
            msg = self.insert_success_msg.format(
                listing=listing, form=form, object=object
            )
            if msg:
                messages.add_message(listing.request, messages.SUCCESS, mark_safe(msg))
        else:
            msg = self.insert_success_msg_no_save.format(
                listing=listing, form=form, object=object
            )
            if msg:
                messages.add_message(listing.request, messages.INFO, mark_safe(msg))

    def send_listing_update_success_message(self, listing, updated_rows_pk):
        nb_updates = len(updated_rows_pk)
        meta = listing.model._meta
        model_verbose = (
            meta.verbose_name_plural if nb_updates > 0 else meta.verbose_name_plural
        )
        if listing.save_to_database:
            msg = self.update_success_msg.format(
                listing=listing, nb_updates=nb_updates, model_verbose=model_verbose
            )
            if msg:
                messages.add_message(listing.request, messages.SUCCESS, mark_safe(msg))
        else:
            msg = self.update_success_msg_no_save.format(
                listing=listing, nb_updates=nb_updates, model_verbose=model_verbose
            )
            if msg:
                messages.add_message(listing.request, messages.INFO, mark_safe(msg))

    def get_listing_id(self, listing_id, refresh=False):
        """For AJAX purposes : when using multiple listings on the same page
        and having one AJAX request : we need only ONE listing instance :
        the one from the given ID in the request. This avoids to create
        all listing instances when not necessary"""
        if isinstance(listing_id, str):
            instance = self.get_from_listing_instances(listing_id)
            if instance is None or refresh == True:
                method = getattr(
                    self, INSTANCE_METHOD_PREFIX + listing_id.replace("-", "_"), None
                )
                if callable(method):
                    instance = method()
                    self.set_to_listing_instances(listing_id, instance)
                else:
                    instance = self.get_default_listing_instance()
            if instance:
                instance.id = listing_id + "-id"
            return instance
        return None

    def get_listings_instances(self):
        prefix_length = len(INSTANCE_METHOD_PREFIX)
        # find all get_listing_instance_xxx
        for method_name in dir(self.__class__):
            if method_name.startswith(INSTANCE_METHOD_PREFIX):
                method = getattr(self, method_name)
                if callable(method):
                    listing_id = method_name[prefix_length:]
                    if not self.in_listing_instances(listing_id):
                        instance = method()
                        instance.id = (
                            listing_id + "-id"
                        )  # Ensure id is set according to method name + '-id'
                        self.set_to_listing_instances(listing_id, instance)
        self.get_default_listing_instance()  # will update self._listing_instances
        for listing in self.yield_listing_instances():
            listing.request = self.request
            if listing.is_initialized() and not listing.is_render_initialized():
                listing.set_view(self)
                listing.render_init(RequestContext(self.request))

        return self.listing_instances_context()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["request"] = self.request  # no need to add request context processor
        context["posted_listing"] = self.listing
        context["view"] = self
        for cls in self.context_classes:
            context[cls.__name__] = cls
        if self.listing_class:
            context[self.listing_class.__name__] = self.listing_class
        context.update(self.get_listings_instances())
        return context


class ListingView(ListingViewMixin, TemplateView):
    pass
